using Microsoft.Extensions.Configuration;
using System.Text.Json;

namespace View_by_Distance.Rename.Models.Binder;

public class AppSettings
{

    public string? Company { get; set; }
    public bool? ForceIdName { get; set; }
    public int? MaxDegreeOfParallelism { get; set; }
    public int? MaxMinutesDelta { get; set; }
    public bool? RenameUndo { get; set; }
    public string? WorkingDirectoryName { get; set; }

    public override string ToString()
    {
        string result = JsonSerializer.Serialize(this, new JsonSerializerOptions() { WriteIndented = true });
        return result;
    }

    private static Models.AppSettings Get(AppSettings? appSettings)
    {
        Models.AppSettings result;
        if (appSettings?.Company is null)
            throw new NullReferenceException(nameof(appSettings.Company));
        if (appSettings?.ForceIdName is null)
            throw new NullReferenceException(nameof(appSettings.ForceIdName));
        if (appSettings?.MaxDegreeOfParallelism is null)
            throw new NullReferenceException(nameof(appSettings.MaxDegreeOfParallelism));
        if (appSettings?.MaxMinutesDelta is null)
            throw new NullReferenceException(nameof(appSettings.MaxMinutesDelta));
        if (appSettings?.RenameUndo is null)
            throw new NullReferenceException(nameof(appSettings.RenameUndo));
        if (appSettings?.WorkingDirectoryName is null)
            throw new NullReferenceException(nameof(appSettings.WorkingDirectoryName));
        result = new(
            appSettings.Company,
            appSettings.ForceIdName.Value,
            appSettings.MaxDegreeOfParallelism.Value,
            appSettings.MaxMinutesDelta.Value,
            appSettings.RenameUndo.Value,
            appSettings.WorkingDirectoryName
        );
        return result;
    }

    public static Models.AppSettings Get(IConfigurationRoot configurationRoot)
    {
        Models.AppSettings result;
        AppSettings? appSettings = configurationRoot.Get<AppSettings>();
        result = Get(appSettings);
        return result;
    }

}