﻿using CliWrap;
using Microsoft.Extensions.Configuration;
using Phares.Shared;
using Serilog;
using ShellProgressBar;
using View_by_Distance.Rename.Models;
using View_by_Distance.Shared.Models;
using View_by_Distance.Shared.Models.Methods;

namespace View_by_Distance.Rename;

public class Rename
{

    private readonly AppSettings _AppSettings;
    private readonly string _WorkingDirectory;
    private readonly Configuration _Configuration;
    private readonly IsEnvironment _IsEnvironment;
    private readonly IConfigurationRoot _ConfigurationRoot;
    private readonly Property.Models.Configuration _PropertyConfiguration;

    public Rename(List<string> args, IsEnvironment isEnvironment, IConfigurationRoot configurationRoot, AppSettings appSettings, string workingDirectory, bool isSilent, IConsole console)
    {
        if (isSilent)
        { }
        if (console is null)
            throw new NullReferenceException(nameof(console));
        _AppSettings = appSettings;
        _IsEnvironment = isEnvironment;
        _WorkingDirectory = workingDirectory;
        _ConfigurationRoot = configurationRoot;
        ILogger? log = Log.ForContext<Rename>();
        Property.Models.Configuration propertyConfiguration = Property.Models.Binder.Configuration.Get(isEnvironment, configurationRoot);
        Configuration configuration = Models.Binder.Configuration.Get(isEnvironment, configurationRoot, propertyConfiguration);
        _PropertyConfiguration = propertyConfiguration;
        _Configuration = configuration;
        propertyConfiguration.Update();
        log.Information(propertyConfiguration.RootDirectory);
        Verify();
        List<string> lines = RenameFilesInDirectories(log);
        if (lines.Any())
        {
            File.WriteAllLines($"D:/Tmp/Phares/{DateTime.Now.Ticks}.tsv", lines);
            _ = Shared.Models.Stateless.Methods.IPath.DeleteEmptyDirectories(propertyConfiguration.RootDirectory);
        }
    }

    private void Verify()
    {
        if (_AppSettings is null)
            throw new NullReferenceException(nameof(_AppSettings));
        if (_IsEnvironment is null)
            throw new NullReferenceException(nameof(_IsEnvironment));
        if (_Configuration is null)
            throw new NullReferenceException(nameof(_Configuration));
        if (_ConfigurationRoot is null)
            throw new NullReferenceException(nameof(_ConfigurationRoot));
        if (_WorkingDirectory is null)
            throw new NullReferenceException(nameof(_WorkingDirectory));
        if (_PropertyConfiguration is null)
            throw new NullReferenceException(nameof(_PropertyConfiguration));
    }

    private static List<(FileHolder, string, string)> GetRenameUndoToDoCollection(ProgressBar progressBar, string[] files)
    {
        List<(FileHolder, string, string)> results = new();
        string[] lines;
        string fileName;
        string? directory;
        FileHolder fileHolder;
        List<string> distinct = new();
        foreach (string file in files)
        {
            progressBar.Tick();
            directory = Path.GetDirectoryName(file);
            if (string.IsNullOrEmpty(directory))
                continue;
            fileName = Path.GetFileName(file);
            if (!fileName.EndsWith(".id"))
                continue;
            lines = File.ReadAllLines(file);
            if (lines.Length < 2)
                continue;
            if (distinct.Contains(lines[1]))
                continue;
            distinct.Add(lines[1]);
            fileHolder = new(lines[0]);
            results.Add(new(fileHolder, directory, lines[1]));
        }
        return results;
    }

    private List<(FileHolder, string, string)> GetToDoCollection(ProgressBar progressBar, string[] files, bool nefPresent)
    {
        List<(FileHolder, string, string)> results = new();
        int? id;
        int season;
        string? message;
        string checkFile;
        DateTime dateTime;
        bool? isWrongYear;
        string seasonName;
        string? directory;
        TimeSpan? timeSpan;
        string directoryName;
        DateTime?[] dateTimes;
        FileHolder fileHolder;
        string[]? ffmpegFiles;
        bool isIgnoreExtension;
        string? seasonDirectory;
        const string jpg = ".jpg";
        DateTime? minimumDateTime;
        string checkFileExtension;
        DateTime? dateTimeFromName;
        DateTime? dateTimeOriginal;
        const string jpeg = ".jpeg";
        DateTime?[] metadataDateTimes;
        List<string> distinct = new();
        string[] directoryNameSegments;
        bool isValidImageFormatExtension;
        DateTime? metadataDateTimeOriginal;
        bool nameWithoutExtensionIsIdFormat;
        IReadOnlyList<MetadataExtractor.Directory> directories;
        foreach (string file in files)
        {
            progressBar.Tick();
            fileHolder = new(file);
            if (!fileHolder.Exists)
                continue;
            directory = Path.GetDirectoryName(file);
            if (string.IsNullOrEmpty(directory))
                continue;
            if (fileHolder.ExtensionLowered == ".id" || fileHolder.ExtensionLowered == ".lsv" || fileHolder.DirectoryName is null)
                continue;
            if (files.Contains($"{fileHolder.FullName}.id"))
                continue;
            nameWithoutExtensionIsIdFormat = Shared.Models.Stateless.Methods.IProperty.NameWithoutExtensionIsIdFormat(fileHolder);
            if (nameWithoutExtensionIsIdFormat)
                continue;
            dateTimeFromName = Shared.Models.Stateless.Methods.IProperty.GetDateTimeFromName(fileHolder);
            isValidImageFormatExtension = _PropertyConfiguration.ValidImageFormatExtensions.Contains(fileHolder.ExtensionLowered);
            isIgnoreExtension = isValidImageFormatExtension && _PropertyConfiguration.IgnoreExtensions.Contains(fileHolder.ExtensionLowered);
            if (!isIgnoreExtension && isValidImageFormatExtension)
                ffmpegFiles = null;
            else
            {
                try
                { directories = MetadataExtractor.ImageMetadataReader.ReadMetadata(file); }
                catch (Exception) { continue; }
                CommandTask<CommandResult> result = Cli.Wrap("ffmpeg.exe")
                    // .WithArguments(new[] { "-ss", "00:00:00", "-t", "00:00:00", "-i", file, "-qscale:v", "2", "-r", "0.01", $"{fileHolder.Name}-%4d.jpg" })
                    .WithArguments(new[] { "-i", file, "-vframes", "1", $"{fileHolder.Name}-%4d.jpg" })
                    .WithWorkingDirectory(fileHolder.DirectoryName)
                    .ExecuteAsync();
                result.Task.Wait();
                ffmpegFiles = Directory.GetFiles(fileHolder.DirectoryName, $"{fileHolder.Name}-*.jpg", SearchOption.TopDirectoryOnly);
                if (!ffmpegFiles.Any())
                    continue;
                fileHolder = new(ffmpegFiles.First());
                if (!fileHolder.Name.EndsWith("-0001.jpg"))
                    throw new Exception();
                isValidImageFormatExtension = _PropertyConfiguration.ValidImageFormatExtensions.Contains(fileHolder.ExtensionLowered);
                isIgnoreExtension = isValidImageFormatExtension && _PropertyConfiguration.IgnoreExtensions.Contains(fileHolder.ExtensionLowered);
                if (isIgnoreExtension || !isValidImageFormatExtension)
                    continue;
                if (fileHolder.DirectoryName is null)
                    continue;
            }
            if (fileHolder.ExtensionLowered == jpeg)
            {
                if (!isIgnoreExtension && isValidImageFormatExtension)
                {
                    if (File.Exists($"{fileHolder.FullName}.id"))
                    {
                        checkFile = Path.Combine(fileHolder.DirectoryName, $"{fileHolder.NameWithoutExtension}{jpg}.id");
                        if (File.Exists(checkFile))
                            continue;
                        if (distinct.Contains(checkFile))
                            continue;
                        distinct.Add(checkFile);
                        results.Add(new(new($"{fileHolder.FullName}.id"), directory, checkFile));
                    }
                    checkFile = Path.Combine(fileHolder.DirectoryName, $"{fileHolder.NameWithoutExtension}{jpg}");
                    if (File.Exists(checkFile))
                        continue;
                    if (distinct.Contains(checkFile))
                        continue;
                    distinct.Add(checkFile);
                    results.Add(new(fileHolder, directory, checkFile));
                    if (nefPresent)
                        results.Add(new(new($"{fileHolder.FullName[..^4]}.tif"), directory, $"{checkFile[..^4]}.tif"));
                    if (nefPresent)
                        results.Add(new(new($"{fileHolder.FullName[..^4]}.nef"), directory, $"{checkFile[..^4]}.nef"));
                    if (File.Exists(checkFile))
                        continue;
                    File.Move(fileHolder.FullName, checkFile);
                    fileHolder = new(checkFile);
                    if (fileHolder.DirectoryName is null)
                        continue;
                }
            }
            (dateTimeOriginal, dateTimes, id, message) = Shared.Models.Stateless.Methods.IProperty.Get(fileHolder, isIgnoreExtension, isValidImageFormatExtension, _PropertyConfiguration.PopulatePropertyId);
            if (ffmpegFiles is not null)
            {
                fileHolder = new(file);
                foreach (string ffmpegFile in ffmpegFiles)
                    File.Delete(ffmpegFile);
            }
            if (fileHolder.DirectoryName is null)
                continue;
            minimumDateTime = dateTimes.Min();
            if (minimumDateTime is null)
                throw new NotSupportedException();
            if (dateTimeOriginal is null)
                timeSpan = null;
            else
                timeSpan = new(Math.Abs(minimumDateTime.Value.Ticks - dateTimeOriginal.Value.Ticks));
            if (timeSpan is not null && timeSpan.Value.TotalMinutes < _AppSettings.MaxMinutesDelta)
                (metadataDateTimeOriginal, metadataDateTimes) = (null, Array.Empty<DateTime?>());
            else
            {
                if (_PropertyConfiguration.IgnoreExtensions.Contains(fileHolder.ExtensionLowered))
                    continue;
                try
                { directories = MetadataExtractor.ImageMetadataReader.ReadMetadata(fileHolder.FullName); }
                catch (Exception) { continue; }
                (metadataDateTimeOriginal, metadataDateTimes) = Metadata.Models.Stateless.Methods.IMetadata.GetDateTimes(fileHolder, directories);
                dateTimeOriginal ??= metadataDateTimeOriginal;
                if (ffmpegFiles is not null && dateTimeOriginal is not null)
                    minimumDateTime = dateTimeOriginal.Value;
                if (dateTimeOriginal is null)
                    timeSpan = null;
                else
                    timeSpan = new(Math.Abs(minimumDateTime.Value.Ticks - dateTimeOriginal.Value.Ticks));
            }
            if (timeSpan is null || timeSpan.Value.TotalMinutes > _AppSettings.MaxMinutesDelta)
                (isWrongYear, seasonDirectory) = (null, !_AppSettings.ForceIdName ? null : Path.Combine(fileHolder.DirectoryName, "Unknown"));
            else
            {
                directoryName = Path.GetFileName(fileHolder.DirectoryName);
                directoryNameSegments = directoryName.Split(' ');
                if (dateTimeFromName is null)
                    isWrongYear = null;
                else
                    (isWrongYear, _) = Shared.Models.Stateless.Methods.IProperty.IsWrongYear(directoryNameSegments, dateTimeFromName.Value.ToString("yyyy"));
                dateTime = minimumDateTime.Value.AddTicks(timeSpan.Value.Ticks);
                (season, seasonName) = Shared.Models.Stateless.Methods.IProperty.GetSeason(dateTime.DayOfYear);
                seasonDirectory = Path.Combine(fileHolder.DirectoryName, $"{dateTime.Year}.{season} {seasonName}");
            }
            if (seasonDirectory is null || (isWrongYear is not null && isWrongYear.Value))
            {
                if (dateTimeFromName is not null && isWrongYear is not null && isWrongYear.Value)
                    minimumDateTime = dateTimeFromName.Value;
                else if (dateTimeOriginal is not null)
                    minimumDateTime = dateTimeOriginal.Value;
                else if (metadataDateTimeOriginal is not null)
                    minimumDateTime = metadataDateTimeOriginal.Value;
                else
                    minimumDateTime = new DateTime?[] { dateTimes.Where(l => l is not null).Min(), metadataDateTimes.Where(l => l is not null).Min() }.Min();
                if (minimumDateTime is null)
                    continue;
                checkFileExtension = fileHolder.ExtensionLowered == jpeg ? jpg : fileHolder.ExtensionLowered;
                checkFile = Path.Combine(fileHolder.DirectoryName, $"{minimumDateTime.Value:yyyy-MM-dd}.{minimumDateTime.Value.Ticks}.{fileHolder.Length}{checkFileExtension}");
                if (checkFile == fileHolder.FullName)
                    continue;
                if (distinct.Contains(checkFile))
                    continue;
                distinct.Add(checkFile);
                results.Add(new(fileHolder, directory, checkFile));
                if (nefPresent)
                    results.Add(new(new($"{fileHolder.FullName[..^4]}.tif"), directory, $"{checkFile[..^4]}.tif"));
                if (nefPresent)
                    results.Add(new(new($"{fileHolder.FullName[..^4]}.nef"), directory, $"{checkFile[..^4]}.nef"));
            }
            else
            {
                if (id is null)
                    continue;
                checkFileExtension = fileHolder.ExtensionLowered == jpeg ? jpg : fileHolder.ExtensionLowered;
                checkFile = Path.Combine(seasonDirectory, $"{id.Value}{checkFileExtension}");
                if (checkFile == fileHolder.FullName)
                    continue;
                if (File.Exists(checkFile))
                {
                    checkFile = string.Concat(checkFile, ".del");
                    if (File.Exists(checkFile))
                        continue;
                }
                if (distinct.Contains(checkFile))
                    continue;
                distinct.Add(checkFile);
                results.Add(new(fileHolder, seasonDirectory, checkFile));
                if (nefPresent)
                    results.Add(new(new($"{fileHolder.FullName[..^4]}.tif"), seasonDirectory, $"{checkFile[..^4]}.tif"));
                if (nefPresent)
                    results.Add(new(new($"{fileHolder.FullName[..^4]}.nef"), seasonDirectory, $"{checkFile[..^4]}.nef"));
            }
        }
        return results;
    }

    private static List<string> Move(ProgressBar progressBar, List<(FileHolder, string)> verifiedToDoCollection)
    {
        List<string> results = new();
        foreach ((FileHolder fileHolder, string to) in verifiedToDoCollection)
        {
            progressBar.Tick();
            results.Add(fileHolder.NameWithoutExtension);
            try
            { File.Move(fileHolder.FullName, to); }
            catch (Exception) { }
        }
        return results;
    }

    private List<string> RenameFilesInDirectories(ILogger log)
    {
        List<string> results = new();
        string message;
        bool nefPresent;
        ProgressBar progressBar;
        const string fileSearchFilter = "*";
        const string directorySearchFilter = "*";
        List<string> distinctDirectories = new();
        List<(FileHolder, string)> verifiedToDoCollection = new();
        List<(FileHolder, string, string)> toDoCollection = new();
        ProgressBarOptions options = new() { ProgressCharacter = '─', ProgressBarOnBottom = true, DisableBottomPercentage = true };
        List<string[]> filesCollection = Shared.Models.Stateless.Methods.IDirectory.GetFilesCollection(_PropertyConfiguration.RootDirectory, directorySearchFilter, fileSearchFilter);
        int count = filesCollection.Select(l => l.Length).Sum();
        foreach (string[] files in filesCollection)
        {
            if (!files.Any())
                continue;
            // foreach (string file in files)
            // {
            //     if (!file.EndsWith(".del"))
            //         continue;
            //     File.Move(file, file[..^4]);
            // }
            // continue;
            distinctDirectories.Clear();
            message = $") Renaming files for <{files.FirstOrDefault()}>";
            progressBar = new(files.Length, message, options);
            if (_AppSettings.RenameUndo)
                toDoCollection.AddRange(GetRenameUndoToDoCollection(progressBar, files));
            else
            {
                nefPresent = files.Any(l => l.EndsWith(".NEF"));
                if (!nefPresent)
                    toDoCollection.AddRange(GetToDoCollection(progressBar, files, nefPresent));
                else
                    toDoCollection.AddRange(GetToDoCollection(progressBar, (from l in files where l.EndsWith(".JPG") select l).ToArray(), nefPresent));
            }
            progressBar.Dispose();
        }
        foreach ((FileHolder fileHolder, string directory, string to) in toDoCollection)
        {
            if (distinctDirectories.Contains(directory))
                continue;
            distinctDirectories.Add(directory);
        }
        foreach (string distinctDirectory in distinctDirectories)
        {
            if (!Directory.Exists(distinctDirectory))
                _ = Directory.CreateDirectory(distinctDirectory);
        }
        foreach ((FileHolder fileHolder, string directory, string to) in toDoCollection)
        {
            if (File.Exists(to))
                continue;
            verifiedToDoCollection.Add(new(fileHolder, to));
            File.WriteAllText($"{to}.id", $"{to}{Environment.NewLine}{fileHolder.FullName}");
        }
        ConsoleKey? consoleKey = null;
        log.Information($"Ready to Move {verifiedToDoCollection.Count} file(s)?");
        for (int y = 0; y < int.MaxValue; y++)
        {
            log.Information("Press \"Y\" key to move file(s), \"N\" key to log file(s) or close console to not move files");
            consoleKey = System.Console.ReadKey().Key;
            if (consoleKey is ConsoleKey.Y or ConsoleKey.N)
                break;
        }
        log.Information(". . .");
        if (consoleKey is null || consoleKey.Value != ConsoleKey.Y)
            log.Information("Nothing moved!");
        else
        {
            message = ") Renaming files";
            progressBar = new(count, message, options);
            results.AddRange(Move(progressBar, verifiedToDoCollection));
            progressBar.Dispose();
            log.Information("Done Moving");
        }
        return results;
    }

}